"""Basic test for validation code can run on every os.

test.png: https://www.pexels.com/fr-fr/photo/route-en-beton-entre-les-arbres-1563356/
"""

import io
import pathlib

import numpy as np
from PIL import Image

from shrinkix import Shrinkix


def test_shrink_input_path(image_path: pathlib.Path) -> None:
    """Check if image is shrink."""
    shrinker = Shrinkix(max_width=512, max_height=512)
    with io.BytesIO() as stream:
        shrinker.shrink(image_path, output=stream)
        assert image_path.stat().st_size > len(stream.getbuffer())


def test_shrink_output_path(
    image_path: pathlib.Path,
    tmp_path: pathlib.Path,
) -> None:
    """Check if image is shrink."""
    shrinker = Shrinkix(max_width=512, max_height=512)
    with io.BytesIO() as stream:
        shrinker.shrink(image_path, output=tmp_path / image_path.name)
        assert image_path.stat().st_size > len(stream.getbuffer())


def test_shrink_input_array(image_path: pathlib.Path) -> None:
    """Check if image is shrink."""
    shrinker = Shrinkix(max_width=512, max_height=512)
    with io.BytesIO() as stream:
        im = Image.open(image_path)
        arr = np.asarray(im)
        shrinker.shrink(arr, output=stream, format="JPEG")
        assert image_path.stat().st_size > len(stream.getbuffer())


def test_shrink_input_stream(image_path: pathlib.Path) -> None:
    """Check if image is shrink."""
    shrinker = Shrinkix(max_width=512, max_height=512)
    with io.BytesIO() as stream:
        shrinker.shrink(image_path.open("rb"), output=stream, format="JPEG")
        assert image_path.stat().st_size > len(stream.getbuffer())



"""Main module."""

from .cli import entrypoint
from .info import (
    __author__,
    __email__,
    __summary__,
    __version__,
)
from .shrinker import Shrinkix

__all__ = [
    "Shrinkix",
    "__author__",
    "__email__",
    "__summary__",
    "__version__",
    "entrypoint",
]



"""Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys
from typing import Optional

import shrinkix as info

sys.path.insert(0, os.path.abspath(".."))

project = info.__name__
copyright = f"{info.__author__} <{info.__email__}>"
author = info.__author__
version = info.__version__
release = info.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "sphinxcontrib.mermaid",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "**tests**"]
todo_include_todos = False
if os.path.exists("resources/favicon.png"):  # noqa: PTH110
    html_favicon: Optional[str] = "resources/favicon.png"
elif os.path.exists("resources/favicon.jpg"):  # noqa: PTH110
    html_favicon = "resources/favicon.jpg"
else:
    html_favicon = None
# If no docstring, inherit from base class
autodoc_inherit_docstrings = True
# Remove 'view source code' from top of page (for html, not python)
html_show_sourcelink = False
# Enable 'expensive' imports for sphinx_autodoc_typehints
set_type_checking_flag = True

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
autodoc_typehints = "description"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["resources"]



"""Module holding metadata."""

from importlib.metadata import Distribution

_DISTRIBUTION = Distribution.from_name(
    "shrinkix",
)
_METADATA = _DISTRIBUTION.metadata

if "Author" in _METADATA:
    __author__ = str(_METADATA["Author"])
    __email__ = str(_METADATA["Author-email"])
else:
    __author__, __email__ = _METADATA["Author-email"][:-1].split(" <", 1)
__version__ = _METADATA["Version"]
__summary__ = _METADATA["Summary"]
__copyright__ = f"{__author__} <{__email__}>"
__issues__ = "https://github.com/Dashstrom/shrinkix/issues"



"""Configuration for all tests."""

import pathlib

import pytest


@pytest.fixture
def resources() -> pathlib.Path:
    """Open test image."""
    return pathlib.Path(__file__).parent.resolve() / "tests" / "resources"


@pytest.fixture
def image_path(resources: pathlib.Path) -> pathlib.Path:
    """Open test image."""
    return resources / "test.jpg"



"""Test for command line interface."""

import contextlib
import io
import os
import subprocess
import sys

import pytest

from shrinkix import entrypoint


def test_cli_version() -> None:
    """Test if the command line interface is installed correctly."""
    name = "shrinkix"
    env = os.environ.get("VIRTUAL_ENV", "")
    if env:
        if os.name == "nt":
            exe = f"{env}\\\\Scripts\\\\{name}.cmd"
            if not os.path.exists(exe):  # noqa: PTH110
                exe = f"{env}\\\\Scripts\\\\{name}.exe"
        else:
            exe = f"{env}/bin/{name}"
    else:
        exe = name
    out = subprocess.check_output((exe, "--version"), text=True, shell=False)
    assert "version" in out
    out = subprocess.check_output(
        (
            sys.executable,
            "-m",
            "shrinkix",
            "--version",
        ),
        text=True,
        shell=False,
    )
    assert "version" in out
    stdout = io.StringIO()
    with contextlib.redirect_stdout(stdout), pytest.raises(SystemExit):
        entrypoint(("--version",))
    assert "version" in stdout.getvalue()


def test_import() -> None:
    """Test if module entrypoint has correct imports."""
    import shrinkix.__main__  # NoQA: F401



"""Module for shrink images."""

import logging
import pathlib
from math import floor, sqrt
from time import time
from typing import Any, Optional

from PIL import Image
from tqdm import tqdm

from .utils import (
    AllImageSource,
    Formats,
    PathLike,
    PathOrFile,
    open_image,
    ratio,
    size,
    verify_format,
)

MAX_COLORS = 256
MAX_SAMPLE = 100_00

logger = logging.getLogger(__name__)


class Shrinkix:
    def __init__(  # noqa: PLR0913
        self,
        *,
        keep_metadata: Optional[bool] = None,
        max_width: Optional[int] = None,
        max_height: Optional[int] = None,
        experimental_color_reduction: Optional[bool] = None,
        verbose: bool = False,
        artist: Optional[str] = None,
        copyright: Optional[str] = None,  # noqa: A002
        background: Optional[str] = None,
        quality: Optional[int] = None,
    ) -> None:
        """Instantiate Shrinkix."""
        self.keep_metadata = keep_metadata is True
        self.experimental_color_reduction = bool(experimental_color_reduction)
        self.max_width = max_width
        self.max_height = max_height
        self.verbose = verbose
        self.copyright = copyright
        self.artist = artist
        self.background = background
        self.quality = int(quality) if quality is not None else None

    def shrink(  # noqa: PLR0912, C901, PLR0915
        self,
        image: AllImageSource,
        output: PathOrFile,
        format: Optional[Formats] = None,  # noqa: A002
        colors: Optional[int] = None,
    ) -> None:
        """Shrink an image."""
        # Get the output format
        if format is None:
            if isinstance(output, (str, pathlib.Path)):
                format = verify_format(pathlib.Path(output).suffix)  # noqa: A001
            elif isinstance(image, (str, pathlib.Path)):
                format = verify_format(pathlib.Path(image).suffix)  # noqa: A001
            else:
                msg = (
                    "Cannot infer the format from the output; "
                    "please specify the format parameter."
                )
                raise ValueError(msg)

        # Load image
        im = open_image(image)

        # Find th best ratio
        w, h = im.size
        ratio = 1.0
        if self.max_width is not None and w > self.max_width:
            ratio = min(self.max_width / w, ratio)
        if self.max_height is not None and h > self.max_height:
            ratio = min(self.max_height / h, ratio)
        if ratio != 1.0:
            im = im.resize(
                (floor(w * ratio), floor(h * ratio)),
                resample=Image.Resampling.NEAREST,
            )

        # Replace background
        if self.background is not None:
            new_im = Image.new("RGBA", im.size, self.background)
            new_im.paste(im, (0, 0), im.convert("RGBA"))
            im = new_im.convert("RGB")

        # Remove metadata
        if not self.keep_metadata:
            data = list(im.convert("RGBA").getdata())
            im = Image.new("RGBA", im.size)
            im.putdata(data)

        # Reduce colors (Palette is not supported on JPEG or WEBP)
        if format not in ("JPEG", "WEBP"):
            im = self.reduce(im, colors=colors)
        else:
            im = im.convert("RGB") if im.mode != "RGB" else im

        # Specify format in options
        options: dict[str, Any] = {"format": format}

        # Add exif information
        import piexif

        if self.artist is not None or self.copyright is not None:
            exif_dict: dict[str, Any] = {"0th": {}}
            if self.artist is not None:
                exif_dict["0th"][piexif.ImageIFD.Artist] = self.artist
            if self.copyright is not None:
                exif_dict["0th"][piexif.ImageIFD.Copyright] = self.copyright
            options["exif"] = piexif.dump(exif_dict)

        # Save with optimization
        if self.quality is None:
            quality = floor(30 + 65 * (1 - min(sqrt(w * h) / 4096, 1)))
            logger.info("Resolved quality is %s", quality)
        else:
            quality = self.quality
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#png
        if format == "PNG":
            options["optimize"] = True
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#jpeg
        elif format == "JPEG":
            options["quality"] = quality
        # https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#webp
        elif format == "WEBP":
            options["lossless"] = False
            options["quality"] = quality
            options["alpha_quality"] = quality
            options["method"] = 6
        if isinstance(output, (str, pathlib.PurePath)):
            output_path = pathlib.Path(output)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            im.save(output_path, **options)
        else:
            im.save(output, **options)

    def export_name(self, path: pathlib.Path, format: Optional[str]) -> str:  # noqa: A002
        """Export name."""
        if format:
            return path.with_suffix(f".{format.lower()}").name
        return path.name

    def bulk(  # noqa: PLR0912
        self,
        files: list[PathLike],
        output: Optional[PathLike],
        inplace: Optional[bool] = None,
        format: Optional[Formats] = None,  # noqa: A002
        colors: Optional[int] = None,
    ) -> None:
        """Shrink a list of file and export it in output."""
        if inplace is None:
            inplace = False
        if format:
            format = verify_format(format)  # noqa: A001

        if inplace:
            if output is not None:
                error_message = '"output" and "inplace" are mutually exclusive'
                raise ValueError(error_message)
        elif output is None:
            error_message = 'You should provide at least "output" or "inplace"'
            raise ValueError(error_message)
        else:
            output = pathlib.Path(output)

        paths: dict[pathlib.Path, pathlib.Path] = {}
        for file in files:
            path = pathlib.Path(file)
            if path.is_dir():
                source_paths = list(path.glob("**/*"))
                for sub_path in source_paths:
                    if sub_path.is_file():
                        name = self.export_name(sub_path, format)
                        parent = sub_path.parent if inplace else output
                        paths[sub_path] = parent / name  # type: ignore[operator]
            else:
                name = self.export_name(path, format)
                parent = path.parent if inplace else output
                paths[path] = parent / name  # type: ignore[operator]

        with tqdm(paths.items()) as bar:
            for src, dst in bar:
                bar.write(f"Shrinking {src}")
                start = time()
                self.shrink(src, dst, colors=colors)
                end = time()
                elapsed = end - start
                bar.write(
                    f"Export at {dst}, "
                    f"ratio: {ratio(src, dst):.2%}, time: {elapsed:.3f}s, "
                    f"size: {size(src)} to {size(dst)}",
                )
                if inplace and src.resolve() != dst.resolve():
                    src.unlink(missing_ok=True)

    def reduce(
        self,
        image: AllImageSource,
        colors: Optional[int] = None,
    ) -> Image.Image:
        """Reduce image colors."""
        # Open and format for model
        im = open_image(image)

        # No optimization on palette or black and white
        if im.mode in ("L", "LA", "P", "PA"):
            return im

        # Detect if alpha
        if im.mode == "RGBA":
            alpha = True
        elif im.mode == "RGB":
            alpha = False
        else:
            im = im.convert("RGBA")
            alpha = True

        # Convert data to numpy array
        import numpy as np

        arr = np.asarray(im)
        h, w = arr.shape[:2]
        X = arr.reshape(-1, 4 if alpha else 3)  # noqa: N806

        if colors is None:
            if X.shape[0] > MAX_SAMPLE:
                index = np.random.choice(X.shape[0], MAX_SAMPLE, replace=False)  # noqa: NPY002
                sample = X[index]
            else:
                sample = X
            _, block_counts = np.unique(
                sample // 16,
                axis=0,
                return_counts=True,
            )
            colors = len(block_counts)
            colors = min(colors, MAX_COLORS)
            logger.info("Use %s colors", colors)

        if not self.experimental_color_reduction:
            return im.quantize(colors)

        # Create model
        from sklearn.cluster import BisectingKMeans

        kmeans = BisectingKMeans(n_clusters=colors, max_iter=100)

        # Resolve palette indexes
        palette_indexes = kmeans.fit_predict(X)

        # Extract the created palette
        palette = (
            np.array(kmeans.cluster_centers_)
            .round()
            .astype(np.uint8)
            .flatten()
            .tolist()
        )

        # Create blank image
        reduced_image = Image.new("P", (w, h))

        # Add palette
        reduced_image.putpalette(palette, rawmode="RGBA" if alpha else "RGB")

        # Paste the pixel data into the image
        reduced_image.putdata(palette_indexes)

        # return the final result
        return reduced_image



"""Utils functions."""

import pathlib
from typing import IO, Literal, Union

from PIL import Image

PathLike = Union[str, pathlib.Path]
PathOrFile = Union[PathLike, IO[bytes]]
AllImageSource = Union[PathOrFile, Image.Image, Image.SupportsArrayInterface]
Formats = Literal["PNG", "JPEG", "JPG", "WEBP"]
FORMAT_MAPPER: dict[str, Formats] = {
    "png": "PNG",
    "jpeg": "JPEG",
    "jpg": "JPEG",
    "webp": "WEBP",
}


def verify_format(format: str) -> Formats:  # noqa: A002
    """Get normalize format and raise error if not supported."""
    key = format.casefold().replace(".", "")
    if key in FORMAT_MAPPER:
        return FORMAT_MAPPER[key]
    message_error = f"Invalid format {format!r}"
    raise ValueError(message_error)


def open_image(image: AllImageSource) -> Image.Image:
    """Open image from all sources."""
    # Load image
    if isinstance(image, (str, pathlib.PurePath)):
        path = pathlib.Path(image)
        im: Image.Image = Image.open(path.open("rb"))
    elif isinstance(image, Image.Image):
        im = image
    else:
        try:
            im = Image.fromarray(image)  # type: ignore[arg-type]
        except AttributeError:
            im = Image.open(image)  # type: ignore[arg-type]
    return im


def ratio(before: pathlib.Path, after: pathlib.Path) -> float:
    """Return compression ratio."""
    return after.stat().st_size / before.stat().st_size


def size(path: pathlib.Path, suffix: str = "B") -> str:
    """Return the human readable size."""
    num: float = path.stat().st_size
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:  # noqa: PLR2004
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"



"""Module for command line interface."""

import argparse
import logging
import sys
from collections.abc import Sequence
from typing import NoReturn, Optional

from .info import __issues__, __summary__, __version__

LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
logger = logging.getLogger(__name__)


class HelpArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:  # pragma: no cover
        """Handle error from argparse.ArgumentParser."""
        self.print_help(sys.stderr)
        self.exit(2, f"{self.prog}: error: {message}\n")


def get_parser() -> argparse.ArgumentParser:
    """Prepare ArgumentParser."""
    parser = HelpArgumentParser(
        prog="shrinkix",
        description=__summary__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s, version {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="verbose mode, enable INFO and DEBUG messages.",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "-k",
        "--keep-metadata",
        action="store_true",
        help="Keep metadata.",
    )
    parser.add_argument(
        "--max-height",
        type=int,
        help="Maximal height of image.",
    )
    parser.add_argument(
        "--max-width",
        type=int,
        help="Maximal width of image.",
    )
    parser.add_argument(
        "path",
        nargs="+",
        help="Path to images or directories.",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        help="Output for all images.",
    )
    parser.add_argument(
        "-i",
        "--inplace",
        action="store_true",
        help="Replace original images.",
    )
    parser.add_argument(
        "-e",
        "--experimental-color-reduction",
        action="store_true",
        help="Use experimental color reduction algorithm.",
    )
    parser.add_argument(
        "-c",
        "--colors",
        type=int,
        required=False,
        help="Colors t use in color reduction.",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["PNG", "JPEG", "WEBP"],
        help="Export format.",
    )
    parser.add_argument(
        "-b",
        "--background",
        help="Replace transparency with this color.",
    )
    parser.add_argument(
        "-q",
        "--quality",
        help="Image quality (JPEG).",
        type=int,
    )
    parser.add_argument(
        "--copyright",
        help="Copyright to add in exif metadata.",
    )
    parser.add_argument(
        "--artist",
        help="Artist to add in exif metadata.",
    )
    return parser


def setup_logging(verbose: Optional[bool] = None) -> None:
    """Do setup logging."""
    # Setup logging
    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.WARNING,
        format="[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s",
    )


def entrypoint(argv: Optional[Sequence[str]] = None) -> None:
    """Entrypoint for command line interface."""
    try:
        parser = get_parser()
        args = parser.parse_args(argv)
        setup_logging(args.verbose)
        from .shrinker import Shrinkix

        shrinker = Shrinkix(
            max_width=args.max_width,
            max_height=args.max_height,
            keep_metadata=args.keep_metadata,
            experimental_color_reduction=args.experimental_color_reduction,
            copyright=args.copyright,
            artist=args.artist,
            background=args.background,
            quality=args.quality,
        )
        if args.output_dir and args.inplace:
            parser.error(
                '"-o/--output-dir" and "-i/--inplace" are mutually exclusive'
            )
        if not args.output_dir or not args.inplace:
            parser.error(
                "you should provide at least "
                '"-o/--output-dir" or "-i/--inplace"'
            )
        shrinker.bulk(
            args.path,
            output=args.output_dir,
            format=args.format,
            inplace=args.inplace,
            colors=args.colors,
        )
    except Exception as err:  # NoQA: BLE001   # pragma: no cover
        logger.critical("Unexpected error", exc_info=err)
        logger.critical("Please, report this error to %s.", __issues__)
        sys.exit(1)
